'use strict';

const vfs = require('@previewjs/vfs');
const fs = require('fs-extra');
const path = require('path');
const typeAnalyzer = require('@previewjs/type-analyzer');
const ts = require('typescript');
const csf3 = require('@previewjs/csf3');
const serializableValues = require('@previewjs/serializable-values');
const compilerSfc = require('@vue/compiler-sfc');

function analyzeVueComponentFromTemplate(typeAnalyzer$1, absoluteFilePath) {
  const tsFilePath = `${absoluteFilePath}.ts`;
  const resolver = typeAnalyzer$1.analyze([tsFilePath]);
  const sourceFile = resolver.sourceFile(tsFilePath);
  for (const statement of sourceFile?.statements || []) {
    const definedProps = extractDefinePropsFromStatement(resolver, statement);
    if (definedProps) {
      return {
        propsType: definedProps.type,
        types: definedProps.collected
      };
    }
    if (ts.isTypeAliasDeclaration(statement) && statement.name.text === "PJS_Props") {
      const type = resolver.checker.getTypeAtLocation(statement);
      const defineComponentProps = resolver.resolveType(type);
      return {
        propsType: defineComponentProps.type,
        types: defineComponentProps.collected
      };
    }
  }
  return {
    propsType: typeAnalyzer.UNKNOWN_TYPE,
    types: {}
  };
}
function extractDefinePropsFromStatement(resolver, statement) {
  if (ts.isExpressionStatement(statement)) {
    const definedProps = extractDefinePropsFromExpression(
      resolver,
      statement.expression
    );
    if (definedProps) {
      return definedProps;
    }
  }
  if (ts.isVariableStatement(statement)) {
    for (const variableDeclaration of statement.declarationList.declarations) {
      if (!variableDeclaration.initializer) {
        continue;
      }
      const definedProps = extractDefinePropsFromExpression(
        resolver,
        variableDeclaration.initializer
      );
      if (definedProps) {
        return definedProps;
      }
    }
  }
  return null;
}
function extractDefinePropsFromExpression(resolver, expression) {
  if (ts.isBinaryExpression(expression) && expression.operatorToken.kind === ts.SyntaxKind.EqualsToken) {
    return extractDefinePropsFromExpression(resolver, expression.right);
  }
  if (!ts.isCallExpression(expression) || !ts.isIdentifier(expression.expression)) {
    return null;
  }
  const functionName = expression.expression.text;
  if (functionName === "defineProps") {
    const signature = resolver.checker.getResolvedSignature(expression);
    if (signature) {
      return resolver.resolveType(signature.getReturnType());
    }
  }
  if (functionName === "withDefaults") {
    const [firstArgument, secondArgument] = expression.arguments;
    if (!firstArgument || !secondArgument) {
      return null;
    }
    const definePropsType = extractDefinePropsFromExpression(
      resolver,
      firstArgument
    );
    if (!definePropsType) {
      return null;
    }
    const defaultsType = resolver.resolveType(
      resolver.checker.getTypeAtLocation(secondArgument)
    );
    if (defaultsType.type.kind !== "object" || definePropsType.type.kind !== "object") {
      return definePropsType;
    }
    const fieldsWithDefaultValue = new Set(
      Object.keys(defaultsType.type.fields)
    );
    return {
      type: typeAnalyzer.objectType(
        Object.fromEntries(
          Object.entries(definePropsType.type.fields).map(
            ([fieldKey, fieldType]) => [
              fieldKey,
              fieldsWithDefaultValue.has(fieldKey) ? typeAnalyzer.optionalType(fieldType) : fieldType
            ]
          )
        )
      ),
      collected: definePropsType.collected
    };
  }
  return null;
}

function inferComponentNameFromVuePath(filePath) {
  return path.basename(filePath, path.extname(filePath));
}

function extractVueComponents(resolver, absoluteFilePath) {
  const sourceFile = resolver.sourceFile(absoluteFilePath);
  if (!sourceFile) {
    return [];
  }
  const functions = [];
  for (const statement of sourceFile.statements) {
    if (ts.isExportAssignment(statement)) {
      if (ts.isIdentifier(statement.expression)) {
        continue;
      }
      functions.push(["default", statement, statement.expression]);
    } else if (ts.isVariableStatement(statement)) {
      for (const declaration of statement.declarationList.declarations) {
        if (!ts.isIdentifier(declaration.name) || !declaration.initializer) {
          continue;
        }
        functions.push([
          declaration.name.text,
          statement,
          declaration.initializer
        ]);
      }
    } else if (ts.isFunctionDeclaration(statement)) {
      const isDefaultExport = !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.ExportKeyword
      ) && !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.DefaultKeyword
      );
      const name = statement.name?.text;
      if (isDefaultExport || name) {
        functions.push([name || "default", statement, statement]);
      }
    }
  }
  const storiesDefaultComponent = csf3.extractDefaultComponent(sourceFile);
  const resolvedStoriesComponent = storiesDefaultComponent ? csf3.resolveComponent(resolver.checker, storiesDefaultComponent) : null;
  const components = [];
  const nameToExportedName = typeAnalyzer.helpers.extractExportedNames(sourceFile);
  const args = typeAnalyzer.helpers.extractArgs(sourceFile);
  function extractComponentTypeInfo(node, name) {
    const storyArgs = args[name];
    const isExported = name === "default" || !!nameToExportedName[name];
    if (storiesDefaultComponent && storyArgs && isExported) {
      return {
        kind: "story",
        args: {
          start: storyArgs.getStart(),
          end: storyArgs.getEnd(),
          value: serializableValues.parseSerializableValue(storyArgs)
        },
        associatedComponent: resolvedStoriesComponent
      };
    }
    const type = resolver.checker.getTypeAtLocation(node);
    for (const callSignature of type.getCallSignatures()) {
      const returnType = callSignature.getReturnType();
      if (isJsxElement(returnType)) {
        return {
          kind: "component",
          exported: isExported,
          analyze: async () => ({
            propsType: typeAnalyzer.UNKNOWN_TYPE,
            types: {}
          })
        };
      }
      if (isExported && returnType.getProperty("template")) {
        return {
          kind: "story",
          args: null,
          associatedComponent: resolvedStoriesComponent
        };
      }
    }
    return null;
  }
  for (const [name, statement, node] of functions) {
    const info = extractComponentTypeInfo(node, name);
    if (info) {
      components.push({
        absoluteFilePath,
        name,
        offsets: [[statement.getStart(), statement.getEnd()]],
        info
      });
    }
  }
  return [...components, ...csf3.extractCsf3Stories(resolver, sourceFile)].map(
    (c) => ({
      ...c,
      info: c.info.kind === "story" ? {
        kind: "story",
        args: c.info.args,
        associatedComponent: c.info.associatedComponent ? transformVirtualTsVueFile(c.info.associatedComponent) : null
      } : c.info
    })
  );
}
function transformVirtualTsVueFile({
  absoluteFilePath,
  name
}) {
  if (absoluteFilePath.endsWith(".vue.ts")) {
    absoluteFilePath = absoluteFilePath.substring(
      0,
      absoluteFilePath.length - 3
    );
    name = inferComponentNameFromVuePath(absoluteFilePath);
  }
  return { absoluteFilePath, name };
}
const jsxElementTypes = /* @__PURE__ */ new Set(["Element"]);
function isJsxElement(type) {
  if (type.isUnion()) {
    for (const subtype of type.types) {
      if (isJsxElement(subtype)) {
        return true;
      }
    }
  }
  return jsxElementTypes.has(type.symbol?.getEscapedName().toString());
}

function createVueTypeScriptReader(reader) {
  return new VueTypeScriptReader(reader);
}
class VueTypeScriptReader {
  constructor(reader) {
    this.reader = reader;
    this.listeners = new vfs.ReaderListeners();
    reader.listeners.add({
      onChange: (absoluteFilePath, info) => {
        this.listeners.notify(absoluteFilePath, info);
      }
    });
    this.observe = reader.observe?.bind(reader);
  }
  async read(absoluteFilePath) {
    if (absoluteFilePath.endsWith(".vue.ts")) {
      const source = await this.reader.read(
        absoluteFilePath.substr(0, absoluteFilePath.length - 3)
      );
      if (source?.kind !== "file") {
        return null;
      }
      const name = path.basename(absoluteFilePath);
      return {
        kind: "file",
        name,
        realPath: async () => null,
        lastModifiedMillis: () => source.lastModifiedMillis(),
        read: async () => convertToTypeScript(await source.read(), name),
        size: () => source.size()
      };
    }
    return this.reader.read(absoluteFilePath);
  }
  readSync(absoluteFilePath) {
    if (absoluteFilePath.endsWith(".vue.ts")) {
      const source = this.reader.readSync(
        absoluteFilePath.substr(0, absoluteFilePath.length - 3)
      );
      if (source?.kind !== "file") {
        return null;
      }
      const name = path.basename(absoluteFilePath);
      return {
        kind: "file",
        name,
        realPath: () => null,
        lastModifiedMillis: () => source.lastModifiedMillis(),
        read: () => convertToTypeScript(source.read(), name),
        size: () => source.size()
      };
    }
    const entry = this.reader.readSync(absoluteFilePath);
    if (entry?.kind === "file" && (entry.name.endsWith(".jsx") || entry.name.endsWith(".tsx"))) {
      return {
        ...entry,
        read: () => `/// <reference types="@vue/runtime-dom" />
${entry.read()}`
      };
    }
    return entry;
  }
}
function convertToTypeScript(vueTemplateSource, name) {
  const parsed = compilerSfc.parse(vueTemplateSource);
  if (parsed.descriptor.scriptSetup) {
    return `import { defineProps } from '@vue/runtime-core';
${parsed.descriptor.scriptSetup.content}`;
  } else if (parsed.descriptor.script) {
    const lang = parsed.descriptor.script.lang;
    if (lang && !["js", "javascript", "ts", "typescript"].includes(lang)) {
      return "";
    }
    const scriptContent = parsed.descriptor.script.content;
    const sourceFile = ts.createSourceFile(
      `${name}.vue.ts`,
      scriptContent,
      ts.ScriptTarget.Latest
    );
    const transformation = ts.transform(sourceFile, [
      (context) => (sourceFile2) => {
        if (ts.isSourceFile(sourceFile2)) {
          return ts.visitEachChild(
            sourceFile2,
            (node) => {
              if (ts.isExportAssignment(node)) {
                return ts.factory.createVariableStatement(
                  [],
                  ts.factory.createVariableDeclarationList(
                    [
                      ts.factory.createVariableDeclaration(
                        "pjs_component",
                        void 0,
                        void 0,
                        ts.factory.createAsExpression(
                          extractDefineComponentArgument(node.expression),
                          ts.factory.createTypeReferenceNode("const")
                        )
                      )
                    ],
                    ts.NodeFlags.Const
                  )
                );
              }
              return node;
            },
            context
          );
        }
        return sourceFile2;
      }
    ]);
    const transformed = ts.createPrinter().printFile(transformation.transformed[0]);
    transformation.dispose();
    return `
${transformed}

import type { PropType as PJS_PropType } from "@vue/runtime-core";

type PJS_TypeOrUnion<T> = PJS_PropType<T> | ReadonlyArray<PJS_PropType<T>>;
type PJS_OptionalPropType<T> = PJS_TypeOrUnion<T> | {type: PJS_TypeOrUnion<T>; required?: false};
type PJS_RequiredPropType<T> = {type: PJS_TypeOrUnion<T>; required: true};
type PJS_OptionalPropsKeys<T> = {
  [K in keyof T]: T[K] extends PJS_OptionalPropType<any> ? K : never;
}[keyof T];
type PJS_RequiredPropsKeys<T> = {
  [K in keyof T]: T[K] extends PJS_RequiredPropType<any> ? K : never;
}[keyof T];
type PJS_CombinedProps<T> = T extends readonly [...any] ? {
  [K in T[number]]: unknown
} : ({
  [K in PJS_OptionalPropsKeys<T>]?: T[K] extends PJS_OptionalPropType<infer S> ? S : never;
} & {
  [K in PJS_RequiredPropsKeys<T>]: T[K] extends PJS_RequiredPropType<infer S> ? S : never;
});
type PJS_ExtractProps<T> = T extends { props: any } ? PJS_CombinedProps<T['props']> : {}
type PJS_Props = PJS_ExtractProps<typeof pjs_component>;
`;
  } else {
    return "";
  }
}
function extractDefineComponentArgument(node) {
  if (ts.isCallExpression(node) && ts.isIdentifier(node.expression) && node.expression.text === "defineComponent" && node.arguments.length > 0) {
    return node.arguments[0];
  }
  return node;
}

const vue3FrameworkPlugin = {
  isCompatible: async (dependencies) => {
    const version = await dependencies["vue"]?.readInstalledVersion() || await dependencies["nuxt"]?.readInstalledVersion() || await dependencies["nuxt3"]?.readInstalledVersion();
    if (!version) {
      return false;
    }
    return parseInt(version) === 3;
  },
  async create({ rootDirPath }) {
    const { default: createVuePlugin } = await import('@vitejs/plugin-vue');
    const { default: vueJsxPlugin } = await import('@vitejs/plugin-vue-jsx');
    const { Parser } = await import('acorn');
    const previewDirPath = path.resolve(__dirname, "..", "preview");
    return {
      pluginApiVersion: 3,
      name: "@previewjs/plugin-vue3",
      defaultWrapperPath: "__previewjs__/Wrapper.vue",
      previewDirPath,
      transformReader: (reader) => vfs.createStackedReader([
        createVueTypeScriptReader(reader),
        vfs.createFileSystemReader({
          mapping: {
            from: path.join(previewDirPath, "modules"),
            to: path.join(rootDirPath, "node_modules")
          },
          watch: false
        })
      ]),
      detectComponents: async (typeAnalyzer, absoluteFilePaths) => {
        const resolver = typeAnalyzer.analyze(absoluteFilePaths);
        const components = [];
        for (const absoluteFilePath of absoluteFilePaths) {
          if (absoluteFilePath.endsWith(".vue") && await fs.pathExists(absoluteFilePath)) {
            components.push({
              absoluteFilePath,
              name: inferComponentNameFromVuePath(absoluteFilePath),
              offsets: [
                [0, (await fs.readFile(absoluteFilePath, "utf-8")).length]
              ],
              info: {
                kind: "component",
                exported: true,
                analyze: async () => analyzeVueComponentFromTemplate(
                  typeAnalyzer,
                  absoluteFilePath
                )
              }
            });
          } else {
            components.push(
              ...extractVueComponents(resolver, absoluteFilePath)
            );
          }
        }
        return components;
      },
      viteConfig: () => {
        return {
          plugins: [
            createVuePlugin(),
            vueJsxPlugin(),
            {
              name: "previewjs:disable-vue-hmr",
              async transform(code, id) {
                if (!id.endsWith(".vue")) {
                  return null;
                }
                const matchHmr = /import\.meta\.hot\.accept\((.|\n)*\}\);?/m;
                return code.replace(matchHmr, "");
              }
            },
            {
              name: "previewjs:process-define-preview",
              async transform(code) {
                if (!code.includes("const _sfc_main = ")) {
                  return;
                }
                let parsed;
                try {
                  parsed = Parser.parse(code, {
                    ecmaVersion: "latest",
                    sourceType: "module"
                  });
                } catch (e) {
                  return null;
                }
                for (const statement of parsed.body || []) {
                  if (statement.type === "VariableDeclaration") {
                    for (const declaration of statement.declarations) {
                      if (declaration.type === "VariableDeclarator" && declaration.id.type === "Identifier" && declaration.id.name === "_sfc_main" && declaration.init && declaration.init.type === "CallExpression" && declaration.init.callee.type === "Identifier" && declaration.init.callee.name === "_defineComponent" && declaration.init.arguments.length > 0 && declaration.init.arguments[0].type === "ObjectExpression") {
                        const properties = declaration.init.arguments[0].properties;
                        const setupProperty = properties.find(
                          (p) => p.key?.name === "setup" && p.value?.type === "FunctionExpression"
                        );
                        if (!setupProperty) {
                          continue;
                        }
                        const setupBody = setupProperty.value.body.body;
                        const definePreviewsCall = setupBody.find(
                          (s) => s.type === "ExpressionStatement" && s.expression.type === "CallExpression" && s.expression.callee.type === "Identifier" && s.expression.callee.name === "definePreviews"
                        );
                        if (!definePreviewsCall || definePreviewsCall.expression.arguments.length === 0) {
                          continue;
                        }
                        const props = definePreviewsCall.expression.arguments[0];
                        return code + `
                          _sfc_main.previews = ${code.slice(
                          props.start,
                          props.end
                        )};
                          `;
                      }
                    }
                  }
                }
                return null;
              }
            },
            {
              name: "previewjs:optimize-deps",
              config: () => ({
                optimizeDeps: {
                  include: ["vue"]
                }
              })
            }
          ],
          esbuild: {
            banner: `import { h } from 'vue';`,
            jsxFactory: "h"
          },
          resolve: {
            alias: {
              vue: "vue/dist/vue.esm-bundler.js"
            }
          }
        };
      }
    };
  }
};

module.exports = vue3FrameworkPlugin;
